/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "alembic/alembic.pmh"

#define FE_SAA_DEBUG	FALSE

namespace fe
{

BWORD SurfaceAccessibleAbc::load(String a_filename,sp<Catalog> a_spSettings)
{
#if FE_SAA_DEBUG
	feLog("SurfaceAccessibleAbc::load file \"%s\"\n",a_filename.c_str());
#endif

	if(a_spSettings.isValid())
	{
		m_frame=a_spSettings->catalog<Real>("frame");
		m_spFilter=a_spSettings->catalog< sp<Component> >("filter");

		const String options=a_spSettings->catalog<String>("options");

#if FE_SAA_DEBUG
		feLog("SurfaceAccessibleAbc::load options \"%s\"\n",options.c_str());
#endif

		if(!options.empty())
		{
			setOptions(options);
		}
	}
	else
	{
		m_frame=0.0;
		m_spFilter=NULL;
	}

	m_fps= -1.0;

	m_minStep= -1.0;

	if(m_optionMap.find("fps")!=m_optionMap.end())
	{
		m_fps=atof(m_optionMap["fps"].c_str());
	}

	if(m_filename!=a_filename)
	{
		m_filename=a_filename;

		m_archive.reset();
	}

#if FE_SAA_DEBUG
	feLog("SurfaceAccessibleAbc::load frame %.6G\n",m_frame);
#endif

	if(!m_archive.valid())
	{
#if FE_SAA_DEBUG
		feLog("SurfaceAccessibleAbc::load opening archive\n");
#endif

		Alembic::AbcGeom::ErrorHandler::Policy policy=
//				Alembic::AbcGeom::ErrorHandler::kQuietNoopPolicy:
				Alembic::AbcGeom::ErrorHandler::kNoisyNoopPolicy;

#if ALEMBIC_LIBRARY_VERSION < 10500
		Alembic::AbcGeom::IArchive* pArchive=new Alembic::AbcGeom::IArchive(
				Alembic::AbcCoreHDF5::ReadArchive(),
				a_filename.c_str(),policy);

		//* TODO verify that this is properly ref-counted underneath
		m_archive= *pArchive;
		delete pArchive;
#else
		Alembic::AbcCoreFactory::IFactory factory;
		factory.setPolicy(policy);
		m_archive=factory.getArchive(a_filename.c_str());
#endif

	}

	if(!m_archive.valid())
	{
		feLog("SurfaceAccessibleAbc::load failed to read \"%s\"\n",
				a_filename.c_str());

		return FALSE;
	}

#if FE_SAA_DEBUG
		std::string appName;
		std::string libraryVersionString;
		Alembic::Util::uint32_t libraryVersion;
		std::string whenWritten;
		std::string userDescription;

		Alembic::Abc::GetArchiveInfo(m_archive,appName,libraryVersionString,
				libraryVersion,whenWritten,userDescription);

		feLog("  written by    \"%s\"\n",appName.c_str());
		feLog("  using Alembic %d \"%s\"\n",
				libraryVersion,libraryVersionString.c_str());
		feLog("  written on    \"%s\"\n",whenWritten.c_str());
		feLog("  description   \"%s\"\n",userDescription.c_str());
#endif

	Real minTimePerSample= -1.0;

	const I32 samplerCount=m_archive.getNumTimeSamplings();
	for(I32 samplerIndex=0;samplerIndex<samplerCount;samplerIndex++)
	{
		Alembic::AbcGeom::TimeSamplingPtr sampler=
				m_archive.getTimeSampling(samplerIndex);

		Alembic::AbcGeom::TimeSamplingType samplerType=
				sampler->getTimeSamplingType();

		const BWORD isUniform=samplerType.isUniform();
		const Real timePerCycle=samplerType.getTimePerCycle();

#if FE_SAA_DEBUG
		const I32 timeCount=sampler->getNumStoredTimes();
		const BWORD isCyclic=samplerType.isCyclic();
		const BWORD isAcyclic=samplerType.isAcyclic();
		const U32 samplesPerCycle=samplerType.getNumSamplesPerCycle();
		const Real fps=Real(1)/timePerCycle;

		feLog("sampler %d/%d times %d uni %d cyc %d acyc %d"
				" SPC %d TPC %.6G FPS %.6G\n",
				samplerIndex,samplerCount,timeCount,
				isUniform,isCyclic,isAcyclic,
				samplesPerCycle,timePerCycle,fps);
#endif

		if(!isUniform)
		{
			feLog("SurfaceAccessibleAbc::load"
					" non-uniform sampling detected\n");
		}

		if(timePerCycle>1e-3 &&
				(minTimePerSample<0.0 || minTimePerSample>timePerCycle))
		{
			minTimePerSample=timePerCycle;
		}
	}

	m_mode=e_scan;
	m_xformCount=0;
	m_meshCount=0;

	Alembic::AbcGeom::IObject iObj=m_archive.getTop();

	visit(iObj);

#if FE_SAA_DEBUG
	feLog("SurfaceAccessibleAbc::load"
			" scan xformCount %d meshCount %d"
			" minStep %.6G minTimePerSample %.6G\n",
			m_xformCount,m_meshCount,m_minStep,minTimePerSample);
#endif

	//* auto-detected frames per second
	if(m_fps<0.0)
	{
		if(m_minStep>0.0)
		{
			m_fps=Real(1)/m_minStep;
		}
		else if(minTimePerSample>0.0)
		{
			m_fps=Real(1)/minTimePerSample;
		}
		else
		{
			m_fps=Real(1.0/24.0);
		}

#if FE_SAA_DEBUG
		feLog("SurfaceAccessibleAbc::load presuming %.6G fps\n",m_fps);
#endif
	}

	//* NOTE trying to auto-detect intended content
	if(m_meshCount)
	{
		m_mode=e_mesh;

		startMeshes();
		visit(iObj);
		completeMeshes();
	}
	else if(m_xformCount>1)
	{
		m_mode=e_joint;

		startJoints();
		visit(iObj);
		completeJoints();
	}

	return TRUE;
}

void SurfaceAccessibleAbc::visit(Alembic::AbcGeom::IObject a_iObj)
{
	const String path=a_iObj.getFullName().c_str();
	const U32 childCount=a_iObj.getNumChildren();

#if FE_SAA_DEBUG
	feLog("\nobject \"%s\" %d\n",path.c_str(),childCount);
#endif

	setIdentity(m_refWorld);

	Alembic::AbcGeom::ICompoundProperty properties=a_iObj.getProperties();
	String indent="";

	if(m_mode!=e_scan)
	{
		visitProperties(properties,indent);
	}

	//* TODO add targets

	if(Alembic::AbcGeom::IPolyMesh::matches(a_iObj.getHeader()))
	{
#if FE_SAA_DEBUG
		feLog("MESH\n");
#endif
		if(m_mode==e_scan)
		{
			m_meshCount++;
		}
		if(m_mode==e_mesh)
		{
			const I32 existingPointCount=m_spOutputPoint->count();

			const I32 addPointCount=m_pointArray.size();
			m_spOutputPoint->append(addPointCount);

			for(I32 index=0;index<addPointCount;index++)
			{
				const I32 pointIndex=existingPointCount+index;
				m_spOutputPoint->set(pointIndex,m_pointArray[index]);
			}

			const I32 existingFaceCount=m_spOutputVertices->count();

			const I32 addFaceCount=m_vertCountArray.size();

			m_primVerts.resize(addFaceCount);
			for(I32 index=0;index<addFaceCount;index++)
			{
				m_primVerts[index].resize(m_vertCountArray[index]);
			}

			I32 vertIndex=0;

			const I32 addVertexCount=m_vertIndexArray.size();

			for(I32 faceIndex=0;faceIndex<addFaceCount;faceIndex++)
			{
				Array<I32>& rVerts=m_primVerts[faceIndex];
				U32 vertexCount=rVerts.size();

				for(U32 vertexIndex=0;vertexIndex<vertexCount;vertexIndex++)
				{
					if(vertIndex>=addVertexCount)
					{
						feLog("SurfaceAccessibleAbc::visit"
								" face counts exceeds face indices\n");
						faceIndex=addFaceCount;
						break;
					}
					rVerts[vertexIndex]=m_vertIndexArray[vertIndex++];
				}
			}

			m_spOutputVertices->append(m_primVerts);

			for(I32 index=0;index<addFaceCount;index++)
			{
				const I32 primitiveIndex=existingFaceCount+index;

				Array<I32>& rVerts=m_primVerts[index];

				const I32 index0=rVerts[0];
				const I32 index1=rVerts[1];
				const I32 index2=rVerts[2];

				const SpatialVector point0=m_pointArray[index0];
				const SpatialVector point1=m_pointArray[index1];
				const SpatialVector point2=m_pointArray[index2];

				const SpatialVector norm=
						unitSafe(cross(point2-point0,point1-point0));

				m_spOutputNormal->set(primitiveIndex,norm);
			}
		}
	}

	if(Alembic::AbcGeom::IXform::matches(a_iObj.getHeader()))
	{
#if FE_SAA_DEBUG
		feLog("TRANSFORM\n");
#endif
		Alembic::AbcGeom::IXform ixform(a_iObj,
				Alembic::AbcGeom::kWrapExisting);

		const I32 sampleCount=ixform.getSchema().getNumSamples();

		Alembic::AbcGeom::TimeSamplingPtr sampler=
				ixform.getSchema().getTimeSampling();

		if(m_mode==e_scan)
		{
			m_xformCount++;

			if(sampleCount>1)
			{
				const Alembic::AbcGeom::chrono_t time0=
						sampler->getSampleTime(0);
				const Alembic::AbcGeom::chrono_t time1=
						sampler->getSampleTime(1);
				const Real step=time1-time0;

				if(step>1e-3 && (m_minStep<0.0 || m_minStep>step))
				{
					m_minStep=step;
				}
			}
		}

#if FALSE // FE_SAA_DEBUG
		for(I32 sampleIndex=0;sampleIndex<sampleCount;sampleIndex++)
		{
			const Alembic::AbcGeom::chrono_t time=
					sampler->getSampleTime(sampleIndex);
			feLog("sample %d/%d frame %.6G time %.6G\n",
					sampleIndex,sampleCount,time*m_fps,time);
		}
#endif

		if(m_mode==e_joint)
		{
			std::pair<Alembic::AbcGeom::index_t,Alembic::AbcGeom::chrono_t>
					nearIndex=sampler->getNearIndex(m_frame/m_fps,sampleCount);

			I32 sampleIndex=nearIndex.first;
#if FE_SAA_DEBUG
			feLog("sampleIndex %d/%d near %.6G\n",
					sampleIndex,sampleCount,m_frame/m_fps);
#endif
			FEASSERT(sampleIndex>=0);
			FEASSERT(!sampleCount || sampleIndex<sampleCount);

			Alembic::AbcGeom::XformSample sample;
			ixform.getSchema().get(sample,
					Alembic::Abc::ISampleSelector(
					Alembic::Abc::index_t(sampleIndex)));

			Alembic::AbcGeom::M44d m44d=sample.getMatrix();
			double* pRaw=m44d.getValue();

			SpatialTransform xformRelative;
			set(xformRelative,pRaw);

			const Color color(0.0,0.0,1.0);			//* TODO
			const Color localScale(1.0,1.0,1.0);	//* TODO

			const String name=path.basename();
			const String parentName=path.pathname().basename();

			setJoint(name,parentName,xformRelative,localScale,
					m_refWorld,color);
		}
	}

	for(U32 childIndex=0;childIndex<childCount;childIndex++)
	{
		Alembic::AbcGeom::IObject iChild(a_iObj,
				a_iObj.getChildHeader(childIndex).getName());

#if FE_SAA_DEBUG
		feLog("\nin \"%s\" child %d/%d \"%s\"\n",
				path.c_str(),childIndex,childCount,
				iChild.getName().c_str());
#endif

		visit(iChild);
	}
}

void SurfaceAccessibleAbc::visitProperties(
	Alembic::AbcGeom::ICompoundProperty a_parent,
	String &a_rIndent)
{
#if FE_SAA_DEBUG
	String oldIndent=a_rIndent;
#endif

	for(U32 propIndex=0;propIndex<a_parent.getNumProperties();propIndex++)
	{
		Alembic::AbcGeom::PropertyHeader header=
				a_parent.getPropertyHeader(propIndex);

		if(header.isCompound())
		{
			visitProperty(Alembic::AbcGeom::ICompoundProperty(a_parent,
					header.getName()),a_rIndent);
		}
		else if(header.isScalar())
		{
			visitProperty(Alembic::AbcGeom::IScalarProperty(a_parent,
					header.getName()),a_rIndent);
		}
		else
		{
			FEASSERT(header.isArray());
			visitProperty(Alembic::AbcGeom::IArrayProperty(a_parent,
					header.getName()),a_rIndent);
		}
	}

#if FE_SAA_DEBUG
	a_rIndent=oldIndent;
#endif
}


void SurfaceAccessibleAbc::visitProperty(
	Alembic::AbcGeom::IScalarProperty a_properties,
	String a_indent)
{
	FEASSERT(!a_properties.isArray());

/*
	const Alembic::AbcCoreAbstract::DataType &dataType=
			a_properties.getDataType();
	const Alembic::Util::uint8_t extent=dataType.getExtent();
	Alembic::Util::Dimensions dimensions(extent);
	Alembic::AbcCoreAbstract::ArraySamplePtr samplePtr=
			Alembic::AbcCoreAbstract::AllocateArraySample(dataType,dimensions);
	const U32 sampleCount=a_properties.getNumSamples();
	for(U32 sampleIndex=0;sampleIndex<sampleCount;++sampleIndex)
	{
		a_properties.get(samplePtr->getData(),
				Alembic::Abc::ISampleSelector(
				Alembic::Abc::index_t(sampleIndex)));
		arraySize=samplePtr->size();
	};
*/

#if FE_SAA_DEBUG
	const String interp=
			a_properties.getMetaData().get("interpretation").c_str();

	std::stringstream dtype;
	dtype << a_properties.getDataType();

	feLog("%sscalar \"%s\" interp '%s' type '%s' samples %d\n",
			a_indent.c_str(),a_properties.getName().c_str(),
			interp.c_str(),dtype.str().c_str(),a_properties.getNumSamples());
#endif
}

void SurfaceAccessibleAbc::visitProperty(
	Alembic::AbcGeom::IArrayProperty a_properties,
	String a_indent)
{
	FEASSERT(a_properties.isArray());

	const String propName=a_properties.getName().c_str();

	std::stringstream dtype;
	dtype << a_properties.getDataType();
	const String dataType=dtype.str().c_str();

	Alembic::AbcCoreAbstract::ArraySamplePtr samplePtr;
	const I32 sampleCount = a_properties.getNumSamples();

#if FE_SAA_DEBUG
	const String interp=
			a_properties.getMetaData().get("interpretation").c_str();

	feLog("%sarray \"%s\" interp '%s' type '%s' samples %d\n",
			a_indent.c_str(),propName.c_str(),
			interp.c_str(),dataType.c_str(),sampleCount);
#endif

	Alembic::AbcGeom::TimeSamplingPtr sampler=
			a_properties.getTimeSampling();
	std::pair<Alembic::AbcGeom::index_t,Alembic::AbcGeom::chrono_t>
			nearIndex=sampler->getNearIndex(m_frame/m_fps,sampleCount);

	I32 sampleIndex=nearIndex.first;
#if FE_SAA_DEBUG
	feLog("%s    sampleIndex %d/%d near %.6G\n",
			a_indent.c_str(),sampleIndex,sampleCount,m_frame/m_fps);
#endif
	FEASSERT(sampleIndex>=0);
	FEASSERT(!sampleCount || sampleIndex<sampleCount);

	a_properties.get(samplePtr,
			Alembic::Abc::ISampleSelector(
			Alembic::Abc::index_t(sampleIndex)));

	const U32 arrayCount=samplePtr->size();
#if FE_SAA_DEBUG
	feLog("%s    arrayCount %d\n",a_indent.c_str(),arrayCount);
#endif
	if(!arrayCount)
	{
		return;
	}

	if(dataType=="string")
	{
		std::string* pString=(std::string*)samplePtr->getData();
		const String string=pString[0].c_str();

#if FE_SAA_DEBUG
		for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
		{
			const String nextString=pString[arrayIndex].c_str();
			if(!arrayIndex || string!=nextString)
			{
				feLog("%s    %d/%d \"%s\"\n",a_indent.c_str(),
						arrayIndex,arrayCount,nextString.c_str());
			}
		}
#endif

		if(propName=="worldreference")
		{
			Real val[16];
			sscanf(string.c_str(),
					"%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f",
					&val[0],&val[1],&val[2],&val[3],
					&val[4],&val[5],&val[6],&val[7],
					&val[8],&val[9],&val[10],&val[11],
					&val[12],&val[13],&val[14],&val[15]);

			set(m_refWorld,val);
		}
	}
	else if(dataType=="int32_t")
	{
		I32* i32Array=(I32*)samplePtr->getData();

#if FE_SAA_DEBUG
		I32 minIndex=0;
		I32 maxIndex=0;
		I32 lastInteger=i32Array[0];

		for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
		{
			const I32 nextInteger=i32Array[arrayIndex];
			if((!arrayIndex || nextInteger!=lastInteger) && arrayIndex<8)
			{
				feLog("%s    %d/%d %d\n",a_indent.c_str(),
						arrayIndex,arrayCount,nextInteger);
			}
			lastInteger=nextInteger;

			if(!arrayIndex || minIndex>nextInteger)
			{
				minIndex=nextInteger;
			}
			if(!arrayIndex || maxIndex<nextInteger)
			{
				maxIndex=nextInteger;
			}
		}
		feLog("%s    min %d max %d\n",a_indent.c_str(),
				minIndex,maxIndex);
#endif

		if(propName==".faceCounts")
		{
			m_vertCountArray.resize(arrayCount);
			for(U32 index=0;index<arrayCount;index++)
			{
				m_vertCountArray[index]=i32Array[index];
			}
		}
		if(propName==".faceIndices")
		{
			m_vertIndexArray.resize(arrayCount);
			for(U32 index=0;index<arrayCount;index++)
			{
				m_vertIndexArray[index]=i32Array[index];
			}
		}
	}
	else if(dataType=="float32_t[3]")
	{
		F32* f32Array=(F32*)samplePtr->getData();

		m_pointArray.resize(arrayCount);

		for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
		{
			const F32* pVector=&f32Array[arrayIndex*3];
			const SpatialVector value(pVector[0],pVector[1],pVector[2]);

			if(propName=="P")
			{
				m_pointArray[arrayIndex]=value;
			}

#if FE_SAA_DEBUG
			if(arrayIndex<8)
			{
				feLog("%s    %d/%d %s\n",a_indent.c_str(),
						arrayIndex,arrayCount,c_print(value));
			}
#endif
		}
	}
}

void SurfaceAccessibleAbc::visitProperty(
	Alembic::AbcGeom::ICompoundProperty a_properties,
	String &a_rIndent)
{
#if FE_SAA_DEBUG
	String schema=a_properties.getMetaData().get("schema").c_str();

	feLog("%scompound \"%s\" schema '%s'\n",
			a_rIndent.c_str(),a_properties.getName().c_str(),schema.c_str());

	String oldIndent=a_rIndent;
	a_rIndent+="  ";
#endif

	visitProperties(a_properties,a_rIndent);

#if FE_SAA_DEBUG
	a_rIndent=oldIndent;
#endif
}

} // namespace
