/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "meta/meta.pmh"

#define	FE_OPC_DEBUG	FALSE

namespace fe
{

FE_DL_PUBLIC BWORD OperatorContext::ms_logged=FALSE;
FE_DL_PUBLIC BWORD OperatorContext::ms_loaded=FALSE;
FE_DL_PUBLIC BWORD OperatorContext::ms_registered=FALSE;
FE_DL_PUBLIC sp<Master> OperatorContext::ms_spMaster;

FE_DL_PUBLIC I32 OperatorContext::ms_fontBase= -1;
FE_DL_PUBLIC I32 OperatorContext::ms_fontRange= -1;
FE_DL_PUBLIC XFontStruct* OperatorContext::ms_pFontStruct=NULL;

OperatorContext::OperatorContext(void):
	m_libDefault("fexOperatorDL")
{
	PrefixLog* pPrefixLog=new PrefixLog();
	pPrefixLog->setPrefix("[34mfe[0m  ");

	feLogger.clear(".*");
	feLogger.setLog("stdout",pPrefixLog);
	feLogger.bind(".*", "stdout");

#if FE_OPC_DEBUG
	feLog("OperatorContext\n");
#endif

	m_spMasterHold=ms_spMaster;

#if FE_OPC_DEBUG
	feLog("OperatorContext done\n");
#endif
}

OperatorContext::~OperatorContext(void)
{
#if FE_OPC_DEBUG
	feLog("~OperatorContext\n");
#endif
	m_spMasterHold=NULL;
	m_spScope=NULL;

#if FE_OPC_DEBUG
	if(ms_spMaster.isValid())
	{
		feLog("~OperatorContext Master count is %d\n",ms_spMaster->count());
	}
#endif

	if(ms_spMaster.isValid() && ms_spMaster->count()==1)
	{
		if(ms_fontBase>=0 && ms_fontRange>=0)
		{
			//* HACK OpenGL
#if TRUE
			glDeleteLists(ms_fontBase,ms_fontRange);
#endif
		}

		ms_fontBase= -1;
		ms_fontRange= -1;

		if(ms_pFontStruct)
		{
			Display* pDisplayHandle=XOpenDisplay(NULL);
			if(pDisplayHandle)
			{
				XFreeFont(pDisplayHandle,ms_pFontStruct);
			}
		}

		ms_pFontStruct=NULL;

#if FE_OPC_DEBUG
		feLog("~OperatorContext releasing Master\n");
#endif
		ms_spMaster=NULL;
	}
}

sp<Master> OperatorContext::master(void)
{
#if FE_OPC_DEBUG
	feLog("OperatorContext::master\n");
#endif

	if(!ms_spMaster.isValid())
	{
		ms_spMaster=new Master();
		ms_loaded=FALSE;
	}

	if(!ms_registered)
	{
		feRegisterSegvHandler();
		ms_registered=TRUE;
	}


#if FE_OPC_DEBUG
	feLog("OperatorContext::master %p\n",ms_spMaster.raw())
#endif

	m_spMasterHold=ms_spMaster;
	return ms_spMaster;
}

sp<Scope> OperatorContext::scope(void)
{
	if(!m_spScope.isValid() && master().isValid())
	{
		m_spScope=master()->catalog()->catalogComponent(
				"Scope","OperatorScope");

#if FE_OPC_DEBUG
	feLog("OperatorContext::scope 0x%p\n",m_spScope.raw())
#endif
	}
	return m_spScope;
}

BWORD OperatorContext::loadLibraries(void)
{
#if FE_OPC_DEBUG
	feLog("OperatorContext::loadLibraries\n");
#endif

	if(ms_loaded)
	{
		return FALSE;
	}

	sp<Registry> spRegistry=master()->registry();

	String libs;
	System::getEnvironmentVariable(m_libEnvVar, libs);
	if(libs.empty())
	{
		System::getEnvironmentVariable("FE_OPERATOR_LIBS", libs);
	}
	if(libs.empty())
	{
		libs=m_libDefault;
	}

	BWORD logUsing=FALSE;
	BWORD logMissing=TRUE;
	const String verbose=System::getVerbose();
	if(verbose=="none")
	{
		logMissing=FALSE;
	}
	else if(verbose=="all")
	{
		logUsing=TRUE;
	}

#if FE_OPC_DEBUG
	feLog("OperatorContext::loadLibraries managing \"%s\"\n",libs.c_str());
#endif

	while(!libs.empty())
	{
		String lib=libs.parse("'",":");
		if(lib.empty())
		{
			break;
		}
		try
		{
			Result result=spRegistry->manage(lib);

			if(!ms_logged)
			{
				if(failure(result))
				{
					if(logMissing)
					{
						feLog("missing %s\n",lib.c_str());
					}
				}
				else
				{
					if(logUsing)
					{
						feLog("adding %s\n",lib.c_str());
					}

					libraryLoaded(lib);
				}
			}
#if FE_OPC_DEBUG
			feLog("OperatorContext::loadLibraries manage \"%s\" result=%p\n",
					lib.c_str(),result);
#endif
		}
		catch(...)
		{
			if(logMissing)
			{
				feLog("faulted %s\n",lib.c_str());
			}
		}
	}

	ms_logged=TRUE;
	ms_loaded=TRUE;

	return TRUE;
}

void OperatorContext::initFont(void)
{
#if FE_OPC_DEBUG
	feLog("OperatorContext::initFont\n");
#endif

	//* HACK load font manually

	//* HACK OpenGL
#if TRUE
	const String name="-*-clean-medium-r-*-*-8-*-*-*-*-*-*-*";
	Display* pDisplayHandle=XOpenDisplay(NULL);
	ms_pFontStruct=XLoadQueryFont(pDisplayHandle,name.c_str());
	ms_fontRange=ms_pFontStruct->max_char_or_byte2+1;
	ms_fontBase=glGenLists(ms_fontRange);

	const Font fontId=ms_pFontStruct->fid;
	const U32 first=ms_pFontStruct->min_char_or_byte2;
	const U32 last=ms_pFontStruct->max_char_or_byte2;

	glXUseXFont(fontId,first,last-first+1,ms_fontBase+first);
#endif
}

I32 OperatorContext::fontBase(void)
{
	if(ms_fontBase<0)
	{
		initFont();
	}

	return ms_fontBase;
}

} // namespace
