/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SPO_DEBUG		FALSE

#include "operator/operator.pmh"

using namespace fe;

void SurfaceProxyOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<bool>("fragment")=false;
	catalog<String>("fragment","label")="Fragment";
	catalog<bool>("fragment","joined")=true;
	catalog<String>("fragment","hint")=
			"Associate collections of primitives which have"
			" matching a fragment attribute.";

	catalog<String>("fragmentAttr")="part";
	catalog<String>("fragmentAttr","label")="By Attr";
	catalog<String>("fragmentAttr","enabler")="fragment";
	catalog<String>("fragmentAttr","hint")=
			"Name of string attribute describing membership"
			" in a collection of primitives.";

	catalog<bool>("snap")=false;
	catalog<String>("snap","label")="Snap To Input";
	catalog<String>("snap","hint")=
			"Snap output to nearest existing input point.";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
}

void SurfaceProxyOp::handle(Record& a_rSignal)
{
#if FE_SPO_DEBUG
	feLog("SurfaceProxyOp::handle \"%s\"\n",name().c_str());
#endif

	const String fragmentAttr=catalog<String>("fragmentAttr");
	BWORD fragmented=(!fragmentAttr.empty() && catalog<bool>("fragment"));

	BWORD snap=catalog<bool>("snap");

	String& rSummary=catalog<String>("summary");
	rSummary=fragmented? fragmentAttr: "";

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,spOutputAccessible,
			e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessorI> spInputPoint;
	if(!access(spInputPoint,spInputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spInputVertices;
	if(!access(spInputVertices,spInputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spOutputFragment;

	I32 fragmentCount=1;

	if(fragmented)
	{
		const String inputGroup="";	//* TODO

		spInputVertices->fragmentWith(SurfaceAccessibleI::e_primitive,
				fragmentAttr,inputGroup);
		fragmentCount=spInputVertices->fragmentCount();
		if(!fragmentCount)
		{
			fragmentCount=1;
			fragmented=FALSE;
		}

		access(spOutputFragment,spOutputAccessible,e_primitive,fragmentAttr);
	}

	const I32 outputPrimitiveCount=fragmentCount*4;
	Array< Array<I32> > primVerts(outputPrimitiveCount);
	Array<String> fragmentOfPrimitive(outputPrimitiveCount);

	sp<SurfaceAccessibleI::FilterI> spFilter;

	for(I32 fragmentIndex=0;fragmentIndex<fragmentCount;fragmentIndex++)
	{
		if(interrupted())
		{
			break;
		}

		String fragmentName;
		I32 filterCount;
		if(fragmented)
		{
			fragmentName=spInputVertices->fragment(fragmentIndex);

			if(!spInputVertices->filterWith(fragmentName,spFilter) ||
					spFilter.isNull())
			{
				continue;
			}

			filterCount=spFilter->filterCount();
		}
		else
		{
			filterCount=spInputVertices->count();
		}

		if(!filterCount)
		{
			continue;
		}

		std::set<I32> pointSet;

		SpatialVector base(0.0,0.0,0.0);
		SpatialVector tip(0.0,0.0,0.0);
		Real farthest2=0.0;
		BWORD first=TRUE;

		for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const I32 reIndex=
					fragmented? spFilter->filter(filterIndex): filterIndex;

			const U32 subCount=spInputVertices->subCount(reIndex);
			if(!subCount)
			{
				continue;
			}

			if(first)
			{
				I32 pointIndex=spInputVertices->integer(reIndex,0);

				//* base is the lowest point index of the first primitive
				for(U32 subIndex=1;subIndex<subCount;subIndex++)
				{
					const I32 otherIndex=
							spInputVertices->integer(reIndex,subIndex);
					if(pointIndex>otherIndex)
					{
						pointIndex=otherIndex;
					}
				}

				base=spInputPoint->spatialVector(pointIndex);
				tip=base;
				first=FALSE;
			}

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const SpatialVector point=
						spInputVertices->spatialVector(reIndex,subIndex);
				const Real distance2=magnitudeSquared(point-base);
				if(farthest2<distance2)
				{
					farthest2=distance2;
					tip=point;
				}

				I32 pointIndex=spInputVertices->integer(reIndex,subIndex);
				pointSet.insert(pointIndex);
			}
		}

		SpatialVector sum(0.0,0.0,0.0);
		I32 count=0;

		for(std::set<I32>::iterator it=pointSet.begin();
				it!=pointSet.end();it++)
		{
			const I32 pointIndex= *it;

			sum+=spInputPoint->spatialVector(pointIndex);
			count++;
		}

		const SpatialVector center=sum/Real(count);

		const SpatialVector stem=unitSafe(tip-base);
		SpatialVector aside(0.0,0.0,0.0);
		Real widest2=0.0;

		for(std::set<I32>::iterator it=pointSet.begin();
				it!=pointSet.end();it++)
		{
			const I32 pointIndex= *it;

			const SpatialVector point=
					spInputPoint->spatialVector(pointIndex);
			const SpatialVector away=point-center;
			const Real along=dot(away,stem);
			const SpatialVector apart=away-along*stem;
			const Real wide2=magnitudeSquared(apart);
			if(widest2<wide2)
			{
				widest2=wide2;
				aside=apart;
			}
		}

		const SpatialVector side=center+aside;

		SpatialVector opposition(0.0,0.0,0.0);
		widest2=0.0;

		for(std::set<I32>::iterator it=pointSet.begin();
				it!=pointSet.end();it++)
		{
			const I32 pointIndex= *it;

			const SpatialVector point=
					spInputPoint->spatialVector(pointIndex);

			const SpatialVector away=point-side;
			const Real along=dot(away,stem);
			const SpatialVector apart=away-along*stem;
			const Real wide2=magnitudeSquared(apart);
			if(widest2<wide2)
			{
				widest2=wide2;

				const SpatialVector awayB=side+apart-center;
				const Real alongB=dot(awayB,stem);

				opposition=awayB-alongB*stem;
			}
		}

#if FALSE
		const SpatialVector spine1=base+(tip-base)*0.1;
		const SpatialVector spine2=base+(tip-base)*0.5;
		const SpatialVector spine3=base+(tip-base)*0.9;
#else
		const SpatialVector spine1=center+(base-center)*0.8;
		const SpatialVector spine2=center;
		const SpatialVector spine3=center+(tip-center)*0.8;
#endif

		const SpatialVector shortSide=0.6*aside;
		const SpatialVector shortOppose=0.6*opposition;

		const I32 pointIndexOut=spOutputPoint->append(9);

		spOutputPoint->set(pointIndexOut,base);
		spOutputPoint->set(pointIndexOut+1,spine1+shortOppose);
		spOutputPoint->set(pointIndexOut+2,spine1+shortSide);
		spOutputPoint->set(pointIndexOut+3,spine2+opposition);
		spOutputPoint->set(pointIndexOut+4,spine2);
		spOutputPoint->set(pointIndexOut+5,spine2+aside);
		spOutputPoint->set(pointIndexOut+6,spine3+shortOppose);
		spOutputPoint->set(pointIndexOut+7,spine3+shortSide);
		spOutputPoint->set(pointIndexOut+8,tip);

		I32 primitiveIndex=fragmentIndex*4;

		primVerts[primitiveIndex].resize(4);
		primVerts[primitiveIndex][0]=pointIndexOut+4;
		primVerts[primitiveIndex][1]=pointIndexOut;
		primVerts[primitiveIndex][2]=pointIndexOut+2;
		primVerts[primitiveIndex][3]=pointIndexOut+5;
		fragmentOfPrimitive[primitiveIndex]=fragmentName;
		primitiveIndex++;

		primVerts[primitiveIndex].resize(4);
		primVerts[primitiveIndex][0]=pointIndexOut+1;
		primVerts[primitiveIndex][1]=pointIndexOut;
		primVerts[primitiveIndex][2]=pointIndexOut+4;
		primVerts[primitiveIndex][3]=pointIndexOut+3;
		fragmentOfPrimitive[primitiveIndex]=fragmentName;
		primitiveIndex++;

		primVerts[primitiveIndex].resize(4);
		primVerts[primitiveIndex][0]=pointIndexOut+4;
		primVerts[primitiveIndex][1]=pointIndexOut+5;
		primVerts[primitiveIndex][2]=pointIndexOut+7;
		primVerts[primitiveIndex][3]=pointIndexOut+8;
		fragmentOfPrimitive[primitiveIndex]=fragmentName;
		primitiveIndex++;

		primVerts[primitiveIndex].resize(4);
		primVerts[primitiveIndex][0]=pointIndexOut+6;
		primVerts[primitiveIndex][1]=pointIndexOut+3;
		primVerts[primitiveIndex][2]=pointIndexOut+4;
		primVerts[primitiveIndex][3]=pointIndexOut+8;
		fragmentOfPrimitive[primitiveIndex]=fragmentName;
		primitiveIndex++;

		if(snap)
		{
			for(I32 offset=0;offset<9;offset++)
			{
				const I32 pointIndex=pointIndexOut+offset;
				SpatialVector point=spOutputPoint->spatialVector(pointIndex);

				SpatialVector closestPoint=point;
				Real closestDistance2= -1.0;

				for(std::set<I32>::iterator it=pointSet.begin();
						it!=pointSet.end();it++)
				{
					const I32 pointIndex= *it;

					const SpatialVector other=
							spInputPoint->spatialVector(pointIndex);

					const Real distance2=magnitudeSquared(other-point);

					if(closestDistance2>distance2 || closestDistance2<0.0)
					{
						closestDistance2=distance2;
						closestPoint=other;
					}
				}

				spOutputPoint->set(pointIndex,closestPoint);
			}
		}
	}

	spOutputVertices->append(primVerts);

	if(spOutputFragment.isValid())
	{
		for(I32 primitiveIndex=0;primitiveIndex<outputPrimitiveCount;
				primitiveIndex++)
		{
			spOutputFragment->set(primitiveIndex,
					fragmentOfPrimitive[primitiveIndex]);
		}
	}

#if FE_HGO_DEBUG
	feLog("SurfaceProxyOp::handle done\n");
#endif
}
